;-----------------------------------------------------------------
;
; Name:		eyes.asm
; Title:	Eye circuit services for robot
; Version:	1.0
; Last updated:	2001.03.10
; Target:	AT90S2313
;
;-----------------------------------------------------------------
;
; DESCRIPTION
;
; The eye circuit referred to above is a photodiode operating
; in zero bias mode (cathode to ground) connected to a two stage
; simple transistor amplifier, which is in turn attached to the
; negative input of the comparator. The circuit diagram for this can
; be summarized as follows:
;
; Photodiode cathode to ground, anode through 100nF capacitor to
; base of T1. T1 is a BC639, base also through 150k to collector, 
; emitter to ground, collector to 30k to drive pin. Collector of 
; T1 decoupled to next stage with 100nF capacitor. 
;
; Pins EYEDRIVEL,R and B (PD4-6) supply the power to drive the 
; first stage amplifiers for the photodiodes.
;
; All three first stage outputs are commoned onto a 1k resistor
; into the base of the next stage, T2, also a BC639, base also
; through 30k to collector, collector through 6k8 to positive
; supply. Collector also feeds into inverting input of on-chip
; comparator.
;
; Simple R-C ADC (3k into 10nF) is driven from RCDRIVE (PB2) and
; the junction is connected to the noninverting input of the
; on-chip comparator (RCREAD).
;
; Timer 1 is used in input capture mode to capture a positive edge on
; the comparator.
;
; Pin IREDDRIVE (PB3) is used to drive an IRED (high=on) through
; another pair of transistors (to amplify current) which is then 
; used to take alternate on/off light/dark readings which are then
; totalled as a difference over several cycles.
;
; PUBLISHED FUNCTIONS
;
; reset_eyes
;
; Resets this module. Called on reset. Sets I/O settings, T1
; compare mode and count rate, discharges RC network.
;
;
; update_eyes
;
; Called on a 2kHz counter to actually run the sampling.
;
; The IRED is run at 50% duty during the sampling. Input counts are
; accumulated using successive addition and subtraction to implement
; a simple high-pass filter which resonates at around 1kHz (which is
; not coincidentally the IRED drive frequency).
;
; This filter effects an average over many cycles (64 in this example)
; to help eliminate random noise, which should add and subtract from
; the accumulator evenly.
;
;
; read_eyes
;
; Reads eye states. Returns R0 as a bitvector, with bit 0 indicating
; EyeL state, bit 1 EyeR, and bit 2 EyeB. Whether the eye is "on" or
; not is determined by the thresholds set by the calibration cycle.
;
; Tests suggest that a measurable signal can be obtained from objects
; between 10 and 20cm away from the eye, depending on size, reflectivity,
; etc.
;
;
; calibrate_eyes
;
; Calibrates the eyes over a few cycles, setting the threshold of each
; eye to a sensible value.
;
; For a real-world app, it would be a good idea to run a calibration
; phase to avoid false readings. This should read the eye in a quies-
; cent state (nothing in view) and get estimates of the peak, minimum
; and typical counts from the integrator/filter. A sensible threshold
; can then be established for triggering the device.
;
; 
;-----------------------------------------------------------------

;;;; .include "robotdef.inc"

;-----------------------------------------------------------------
;
; Defines

.equ	EYE_SAMPLES	=$40	; number of samples taken for reading
.equ	EYE_REST		=$10	; number of cycles to wait after switching eyes
.equ	EYE_CALIBRATECOUNT=$40	; number of cycle to calibrate eye

;-----------------------------------------------------------------
;
; Variables

.dseg

eye_count:	.byte	1	; interrupt counter
eye_active:	.byte	1	; active eye
eye_state:	.byte	1	; eye drive mask
eye_ired:	.byte	1	; ired drive state

eye_accuml:	.byte	1	; accumulator for dsp-style filter calculation
eye_accumh:	.byte	1

eye_valuel:	.byte	1	; accumulator is saturated to lie between
eye_valuer:	.byte	1	; +00 and +FF during transfer.
eye_valueb:	.byte	1

eye_thresholdll: .byte	1	; low thresholds set by calibrate_eye
eye_thresholdrl: .byte	1	; these are also used to store the MAXIMUM value
eye_thresholdbl: .byte	1	; encountered during calibration.

eye_thresholdlh: .byte	1	; high thresholds set by calibrate_eye
eye_thresholdrh: .byte	1	; paradoxically, these are also used to store
eye_thresholdbh: .byte	1	; the MINIMUM value encountered during calibration

eyes_triggered:	.byte 	1	; current trigger state for each eye

;-----------------------------------------------------------------
;
; Code
;
; Y is assumed to point to the RAM.
; RAM is assumed to be cleared to 0's on reset.

.cseg

;-----------------------------------------------------------------
;
; reset_eyes

reset_eyes:
	; discharge at start

	ldi	rTEMP,(EYE_SAMPLES+EYE_REST)	; count + recovery time
	sty	eye_count,rTEMP

	ldi	rTEMP,mEYEDRIVEL	; Initial photodiode drive mask
	sty	eye_state,rTEMP

	rcall	eye_drive
		
	; set up interrupt masks to enable timer 1 interrupt for timeout
	; and input capture.

	in	rTEMP,TIMSK
	ori 	rTEMP, (1<<TOIE1)|(1<<TICIE)
	out	TIMSK,rTEMP

	; set up analog comparator mode to allow input capture

	ldi	rTEMP,$07	; ACIC enable, rising edge
	out	ACSR,rTEMP

	ret

;-----------------------------------------------------------------
;
; update_eyes
; Called from within the timer routine on 2kHz timer 0.

update_eyes:
	ldy	rITEMP,eye_count	; tick counter
	dec	rITEMP
	sty	eye_count,rITEMP
	brmi	update_eyes_resetcount

	cpi	rITEMP,EYE_SAMPLES	; still resting?
	brcc	update_eyes_skipcount

	ser	rITEMP		; start count = $FF00
	clr	rITEMP2
	out	TCNT1H,rITEMP
	out	TCNT1L,rITEMP2

	ldi	rITEMP,mDDRBREAD
	out	DDRB,rITEMP	; set data direction for reading
	sbi	PORTB,bRCDRIVE	; drive RC network, all other bits unchanged

	sbic	ACSR,ACO	; already triggered? skip if ok
	rjmp	update_eyes_premature	; premature timeout!

	ldi	rITEMP, $41	; start counter1, no predivider
	out	TCCR1B,rITEMP

update_eyes_skipcount:
	ret

update_eyes_premature:
	ldy	rITEMP2,eye_ired
	ldi	rITEMP,$08	; toggle IRED drive after reading
	eor	rITEMP2,rITEMP
	sty	eye_ired,rITEMP2

	ret

update_eyes_resetcount:
	; Store accumulator value in appropriate eye.

	clr	rITEMP2

	ldy	rISCRATCH,eye_accumh
	and	rISCRATCH,rISCRATCH
	brmi	eye_valuefound

	ser	rITEMP2
	brne	eye_valuefound

	ldy	rITEMP2,eye_accuml

eye_valuefound:
	mov	rISCRATCH2,XL

	ldi	XL,eye_valuel		; address of eye
	ldy	rISCRATCH,eye_active	; +active eye index
	add	XL,rISCRATCH
	st	X,rITEMP2

	mov	XL,rISCRATCH2
	
	ldy	rITEMP,eye_state	; advance to next eye
	cpi	rITEMP,mEYEDRIVEB
	breq	eye_state_reset

	inc	rISCRATCH
	add	rITEMP,rITEMP
	rjmp	eye_next_done

eye_state_reset:
	clr	rISCRATCH
	ldi	rITEMP,mEYEDRIVEL

eye_next_done:
	sty	eye_active,rISCRATCH
	sty	eye_state,rITEMP

	clr	rISCRATCH
	sty	eye_accumh,rISCRATCH
	sty	eye_accuml,rISCRATCH	
	sty	eye_ired,rISCRATCH	

	ldi	rITEMP,(EYE_SAMPLES+EYE_REST)	; count + recovery time
	sty	eye_count,rITEMP
	
	rjmp	eye_drive

;-----------------------------------------------------------------
;
; Timer1 vectors - capture and accumulate value.

eye_overflow:
	in	rISREGSAVE,SREG

	ldi	rITEMP, $40	; stop counter1 - timed out
	out	TCCR1B,rITEMP

	clr	rISCRATCH
	com	rISCRATCH	; $FF
	rjmp	eye_setvalue

eye_capture:
	in	rISREGSAVE,SREG

	ldi	rITEMP, $40	; stop counter1
	out	TCCR1B,rITEMP

	in	rISCRATCH2,ICR1H	; unused, but required for semantics of 16 bit read.
	in	rISCRATCH ,ICR1L

eye_setvalue:
	clr	rISCRATCH2

	ldy	rISCRATCH3,eye_accuml
	ldy	rISCRATCH4,eye_accumh

	ldy	rITEMP2,eye_ired
	and	rITEMP2,rITEMP2
	brne	eye_subtract

	add	rISCRATCH3,rISCRATCH
	adc	rISCRATCH4,rISCRATCH2	; 0
	rjmp	eye_add

eye_subtract:
	sub	rISCRATCH3,rISCRATCH
	sbc	rISCRATCH4,rISCRATCH2	; 0

eye_add:
	rcall	eye_drive

	ldi	rITEMP,mIREDDRIVE	; toggle IRED drive after reading
	eor	rITEMP2,rITEMP		; IRED state still in rITEMP2
	sty	eye_ired,rITEMP2

	sty	eye_accuml,rISCRATCH3
	sty	eye_accumh,rISCRATCH4

	out	SREG,rISREGSAVE
	reti

;-----------------------------------------------------------------
;
; eye_drive
; Common routine to reset ADC RC network
; PD drive value passed in register DRIVE

eye_drive:
	in	rITEMP,PORTD
	cbr	rITEMP,mEYEDRIVES
	ldy	rISCRATCH,eye_state
	or	rITEMP,rISCRATCH
	out	PORTD,rITEMP	; and output drive values

	in	rITEMP,PORTB
	cbr	rITEMP,(mANALOGIN|mRCDRIVE|mIREDDRIVE)
	ldy	rISCRATCH,eye_ired
	or	rITEMP,rISCRATCH
	out	PORTB,rITEMP

	ldi	rITEMP,mDDRDISCHARGE	; outputs on B0, B2, B3-B7
	out	DDRB,rITEMP
	ret

;-----------------------------------------------------------------
;
; read_eyes
; Read eye state and return as bit vector. This uses the low and high
; thresholds, together with the current state, as hysteresis.

read_eyes:
	ldi	rPARAM1,$04		; mask for eye 2 to start
	clr	rRETURN
	ldi	ZL,eye_valuel+3		; one past end of value array
	
read_eyes_loop:
	ld	rSCRATCH,-Z
	ldd	rSCRATCH2,Z+(eye_thresholdll-eye_valuel)	; low threshold

	ldy	rTEMP,eyes_triggered
	and	rTEMP,rPARAM1		; check if this eye triggered already
	brne	read_eyes_hysteresis	; yes - continue to use low threshold
	
	ldd	rSCRATCH2,Z+(eye_thresholdlh-eye_valuel)	; else use high one instead

read_eyes_hysteresis:
	cp	rSCRATCH,rSCRATCH2
	brcs	read_eyes_low
	
	or	rRETURN,rPARAM1		; higher than selected threshold so set

read_eyes_low:
	lsr	rPARAM1			; finished all eyes?
	brne	read_eyes_loop		; not yet
	
	sty	eyes_triggered,rRETURN	; store for next time
	ret

;-----------------------------------------------------------------
;
; calibrate_eyes
; Waits for EYE_CALIBRATECOUNT full eye cycles  and records
; min and max for each eye, which are then used to set the thresholds
; for sensing

calibrate_eyes:
	ldi	rLOOPI,EYE_CALIBRATECOUNT
	
calibrate_eyes_counted:
	clr	rTEMP
	sty	eye_thresholdll,rTEMP	; maxima = minvalue to start with
	sty	eye_thresholdrl,rTEMP
	sty	eye_thresholdbl,rTEMP
	
	ser	rTEMP
	sty	eye_thresholdlh,rTEMP	; minima = maxvalue to start with
	sty	eye_thresholdrh,rTEMP
	sty	eye_thresholdbh,rTEMP

calibrate_eyes_mainloop:	
calibrate_eyes_wait02:
	ldy	rTEMP2,eye_active
	cpi	rTEMP2,$02
	brne	calibrate_eyes_wait02

calibrate_eyes_wait00:
	ldy	rTEMP2,eye_active
	cpi	rTEMP2,$00
	brne	calibrate_eyes_wait00

	; compare each value against min and max and adjust
	; this is done with a predecrement on Z therefore we
	; start z one beyond the end of the value array
	; we can also easily check z for end condition which
	; eliminates the need for a separate loop counter
	
	ldi	ZL,eye_valuel+3
	
calibrate_eyes_minimax:
	ld	rSCRATCH,-Z
	ldd	rSCRATCH2,Z+(eye_thresholdll-eye_valuel)	; max
	ldd	rSCRATCH3,Z+(eye_thresholdlh-eye_valuel)	; min
	
	cp	rSCRATCH,rSCRATCH2
	brcs	calibrate_eyes_notmax
	
	std	Z+(eye_thresholdll-eye_valuel),rSCRATCH

calibrate_eyes_notmax:
	cp	rSCRATCH,rSCRATCH3
	brcc	calibrate_eyes_notmin
	
	std	Z+(eye_thresholdlh-eye_valuel),rSCRATCH

calibrate_eyes_notmin:
	cpi	ZL,eye_valuel
	brne	calibrate_eyes_minimax
	
	dec	rLOOPI
	brne	calibrate_eyes_mainloop
	
	; postprocess min/max into thresholds
	; loop works the same way as for the minimax calculation

	ldi	ZL,eye_thresholdll+3
	
calibrate_eyes_threshold:
	ld	rSCRATCH,-Z		; max
	ldd	rSCRATCH2,Z+(eye_thresholdlh-eye_thresholdll)	; min
	
	sub	rSCRATCH2,rSCRATCH	; min-max = -(max-min)
	
;	Half difference threshold

	asr	rSCRATCH2		; (min-max)/2
	sub	rSCRATCH,rSCRATCH2	; max-(min-max)/2 = max+(max-min)/2

;	Single difference threshold
;
;	sub	rSCRATCH,rSCRATCH2	; max-(min-max) = max+(max-min)

;	Double difference threshold
;
;	sub	rSCRATCH,rSCRATCH2	; max-(min-max) = max+(max-min)
;	sub	rSCRATCH,rSCRATCH2	; max+(max-min)*2
	
	std	Z+(eye_thresholdlh-eye_thresholdll), rSCRATCH	; store back again

	cpi	ZL,eye_thresholdll
	brne	calibrate_eyes_threshold

	ret

;-----------------------------------------------------------------
	