;-----------------------------------------------------------------
;
; Name:		monitor.asm
; Title:	Robot monitor service
; Version:	1.0
; Last updated:	2001.03.19
; Target:	AT90S2313
;
;-----------------------------------------------------------------
;
; DESCRIPTION
;
; Monitor service which runs from the serial port.
; 
; This implements a simple monitor for influencing the behavior
; of the robot and loading eeprom data tables, etc.
;
; Command set
;
; ; 	
;	Comment - rest of line ignored
; 0..7
;	Trigger event 0..7 (6,7 reserved for now)
; d<s>	
;	Display state <s>
; s<s> mm t s s s s s s
;	mm=motor state (0,F,R,?)
;	t=timeout 0-7
;	s=destination state
;
; States are numbered 0-9, A-Z (or a-z)
;
; Execution of the state machine for the robot is suspended
; until completion of a monitor command with trailing crlf
; 
;-----------------------------------------------------------------

.dseg

monitor_select:	.byte	1	; selected state
monitor_sblock:	.byte	4	; state block

;-----------------------------------------------------------------

.cseg

;-----------------------------------------------------------------
;
; reset_monitor
; Reset monitor status. Should be called after reset_uart.

reset_monitor:
	ldi	zl,low(monitor_welcome*2)
	ldi	zh,high(monitor_welcome*2)
monitor_exitmessage:
	rcall	uart_writestringlf
	clr	zh
	ret
	
;-----------------------------------------------------------------
;
; update_monitor
; Main monitor routine.

update_monitor:
	rcall	uart_charready
	brne	update_monitor_charready
	ret

update_monitor_charready:
	rcall	monitor_readchar
	
	cpi	rPARAM1,'0'
	brlt	update_monitor_notdigit
	cpi	rPARAM1,'8'
	brlt	update_monitor_event

update_monitor_notdigit:
	cpi	rPARAM1,$0d
	breq	update_monitor_comment
	cpi	rPARAM1,$0a
	breq	update_monitor_comment
	cpi	rPARAM1,$3B			; semicolon - can't use ';' because it starts a comment. Bleah.
	breq	update_monitor_comment

	cpi	rPARAM1,'d'
	breq	update_monitor_displaystate
	
	cpi	rPARAM1,'s'
	breq	update_monitor_setstate

	ldi	rPARAM1,'*'
	rcall	uart_writechar
	
	rcall	monitor_ignoreline_charread
	rjmp	update_monitor_error
	
update_monitor_comment:
	rcall	monitor_ignoreline_charread
	rjmp	update_monitor_exit
	
update_monitor_event:
	ldi	rTEMP,$01
update_monitor_event_loop:
	cpi	rPARAM1,'0'
	breq	update_monitor_event_set
	lsl	rTEMP
	dec	rPARAM1
	rjmp	update_monitor_event_loop
update_monitor_event_set:
	ldy	rTEMP2,state_events
	or	rTEMP2,rTEMP
	sty	state_events,rTEMP2
	rjmp	update_monitor_exit

update_monitor_displaystate:
	rcall	monitor_readchar
	rcall	monitor_ignoreline	; ignore rest of line before continuing
	
	rcall	monitor_chartostate
	
	rcall	uart_writecrlf
	
update_monitor_showstate:
	mov	zl,rRETURN
	rcall	state_readee		; get motor bits etc.
	
	mov	rTEMP,rRETURN
	ldi	rPARAM1,'0'
	andi	rTEMP,STATE_mMOTORL	; motor bits
	breq	update_monitor_motorl

	ldi	rPARAM1,'R'
	cpi	rTEMP,mMOTORLA
	breq	update_monitor_motorl
		
	ldi	rPARAM1,'F'
	cpi	rTEMP,mMOTORLB
	breq	update_monitor_motorl
	
	ldi	rPARAM1,'?'
update_monitor_motorl:
	rcall	uart_writechar
	
	mov	rTEMP,rRETURN
	ldi	rPARAM1,'0'
	andi	rTEMP,STATE_mMOTORR	; motor bits
	breq	update_monitor_motorr

	ldi	rPARAM1,'R'
	cpi	rTEMP,mMOTORRA
	breq	update_monitor_motorr
		
	ldi	rPARAM1,'F'
	cpi	rTEMP,mMOTORRB
	breq	update_monitor_motorr
	
	ldi	rPARAM1,'?'
update_monitor_motorr:
	rcall	uart_writechar
	
	ldi	rPARAM1,' '
	rcall	uart_writechar
	
	mov	rPARAM1,rRETURN
	rcall	uart_writehexit
	
	ldi	rPARAM1,' '
	rcall	uart_writechar
	
	ldi	rLOOPI,3

update_monitor_ds_loop:
	inc	zl
	rcall	state_readee
	
	mov	rPARAM2,rRETURN
	rcall	update_monitor_display_dest
	swap	rPARAM2
	rcall	update_monitor_display_dest

	dec	rLOOPI
	brne	update_monitor_ds_loop
	
	rjmp	update_monitor_exit



update_monitor_setstate:
	rcall	monitor_readchar	; read state character
	
	rcall	monitor_chartostate	; convert to pointer
	mov	zl,rRETURN
	cpi	zl,EEEND
	brlt	update_monitor_ss_ok1

	;
	; Note: This is interleaved here in order that it's not too far away from
	; the code that uses it.
	;
	
update_monitor_ss_error:	
	ldi	rPARAM1,'*'
	rcall	uart_writechar
	rcall	monitor_ignoreline	; ignore rest of line before continuing
	rjmp	update_monitor_error

update_monitor_ss_ok1:
	sty	monitor_select,zl	; store state pointer
	
	rcall	monitor_readchar	; read first motor character
	clr	rPARAM2			; rPARAM2 = first state block byte
	
	cpi	rPARAM1,'0'
	breq	update_monitor_ss_r

	ori	rPARAM2,mMOTORLA
	cpi	rPARAM1,'R'
	breq	update_monitor_ss_r
	cpi	rPARAM1,'r'
	breq	update_monitor_ss_r

	ori	rPARAM2,mMOTORLB
	cpi	rPARAM1,'?'
	breq	update_monitor_ss_r

	cbr	rPARAM2,mMOTORLA
	cpi	rPARAM1,'F'
	breq	update_monitor_ss_r
	cpi	rPARAM1,'f'
	brne	update_monitor_ss_error

update_monitor_ss_r:
	rcall	monitor_readchar	; read 2nd motor character

	cpi	rPARAM1,'0'
	breq	update_monitor_ss_t

	ori	rPARAM2,mMOTORRA
	cpi	rPARAM1,'R'
	breq	update_monitor_ss_t
	cpi	rPARAM1,'r'
	breq	update_monitor_ss_t

	ori	rPARAM2,mMOTORRB
	cpi	rPARAM1,'?'
	breq	update_monitor_ss_t

	cbr	rPARAM2,mMOTORRA
	cpi	rPARAM1,'F'
	breq	update_monitor_ss_t
	cpi	rPARAM1,'f'
	brne	update_monitor_ss_error

update_monitor_ss_t:
	rcall	monitor_readchar	; read timeout character
	
	subi	rPARAM1,'0'
	brcs	update_monitor_ss_error
	cpi	rPARAM1,8
	brge	update_monitor_ss_error
	
	or	rPARAM2,rPARAM1
	
	ldi	xl,monitor_sblock	; point to state block to fill in
	st	x+,rPARAM2		; store in parameter block
	
	clr	rPARAM2

	ldi	rLOOPI,6		; now read 6 state characters
	
update_monitor_ss_loop:
	rcall	monitor_readchar	; read state character
	
	ldi	rTEMP,$08
	
	cpi	rPARAM1,'-'		; null transition?
	breq	update_monitor_ss_null
	
	rcall	monitor_chartostate	; get destination state
	mov	rTEMP,rRETURN		; get destination state
	ldy	rTEMP2,monitor_select	; get current state
	sub	rTEMP,rTEMP2		; subtract
	asr	rTEMP
	asr	rTEMP			; /4 = state number
	subi	rTEMP,-16		; add 16 to make the compares easier
	cpi	rTEMP,(16-7)
	brlt	update_monitor_ss_error2 ; too far back - error
	cpi	rTEMP,(16+8)
	brlt	update_monitor_ss_ok2

update_monitor_ss_error2:
	rjmp	update_monitor_ss_error	; too far forward - error
	
update_monitor_ss_ok2:
	andi	rTEMP,$0F		; bottom nybble only

update_monitor_ss_null:
	or	rPARAM2,rTEMP		; or in the state
	
	mov	rTEMP,rLOOPI		; get current loop counter
	andi	rTEMP,$01		; odd or even?
	breq	update_monitor_ss_even
	
	st	x+,rPARAM2		; store state

	clr	rPARAM2			; and clear ready for next loop
	
	rjmp	update_monitor_ss_next

update_monitor_ss_even:
	swap	rPARAM2			; swap nybble up into top of state

update_monitor_ss_next:
	dec	rLOOPI
	brne	update_monitor_ss_loop
	
	;
	; Here, we should have a complete state block. Write it to the eeprom.
	;
	
	ldy	zl,monitor_select
	ldi	xl,monitor_sblock
	ldi	rLOOPI,4		; 4 bytes to write to eeprom
	
update_monitor_ss_write:
	ld	rPARAM1,x+
	rcall	monitor_write_eeprom	; write rPARAM1 to eeprom at Z
	inc	zl

	dec	rLOOPI
	brne	update_monitor_ss_write

	rcall	monitor_ignoreline	; ignore rest of line before continuing
	rcall	uart_writecrlf
	rjmp	update_monitor_exit

update_monitor_error:
	ldi	zl,low(monitor_errormsg*2)
	ldi	zh,high(monitor_errormsg*2)
	rjmp	monitor_exitmessage	

update_monitor_exit:
	ldi	zl,low(monitor_exitmsg*2)
	ldi	zh,high(monitor_exitmsg*2)
	rjmp	monitor_exitmessage
	
;-----------------------------------------------------------------
;
; update_monitor_display_dest
; Takes zl = current state, rPARAM2 (high bits) = transition code

update_monitor_display_dest:
	mov	rPARAM1,rPARAM2
	andi	rPARAM1,$F0
	cpi	rPARAM1,$80
	brne	update_monitor_dd_non0
	ldi	rPARAM1,'-'
	rjmp	uart_writechar

update_monitor_dd_non0:
	asr	rPARAM1
	asr	rPARAM1
	add	rPARAM1,zl
	rcall	monitor_statetochar
	rjmp	uart_writechar
	
;-----------------------------------------------------------------
;
; monitor_ignoreline
; Ignore the rest of the line. Call monitor_ignoreline_charread if
; the first character of the rest of the line has already been read.
; Preserves rPARAM1

monitor_ignoreline_charread:
	push	rPARAM1
	rjmp	monitor_ignoreline_charread2

monitor_ignoreline:
	push	rPARAM1
	
monitor_ignoreline_loop:
	rcall	monitor_readchar

monitor_ignoreline_charread2:
	cpi	rPARAM1,$0D
	breq	monitor_ignoreline_exit
	cpi	rPARAM1,$0A
	brne	monitor_ignoreline_loop
	
monitor_ignoreline_exit:
	pop	rPARAM1
	ret

;-----------------------------------------------------------------
;
; monitor_chartostate
; Convert incoming character in rPARAM1 to state pointer in rRETURN

monitor_chartostate:
	cpi	rPARAM1,'A'
	brlt	monitor_chartostate_2

	andi	rPARAM1,$FF-$20		; assume caps
	subi	rPARAM1,7
	
monitor_chartostate_2:
	subi	rPARAM1,'0'
	mov	rRETURN,rPARAM1
	lsl	rRETURN
	lsl	rRETURN
	ret
	
;-----------------------------------------------------------------
;
; monitor_statetochar
; Convert state pointer in rPARAM1 to character in rPARAM1

monitor_statetochar:
	lsr	rPARAM1
	lsr	rPARAM1
	
	cpi	rPARAM1,10
	brlt	monitor_statetochar_2
	
	subi	rPARAM1,-7
	
monitor_statetochar_2:
	subi	rPARAM1,-'0'
	ret

;-----------------------------------------------------------------
;
; monitor_readchar
; Read character with echo back to host. Character is in BOTH rPARAM1
; and rRETURN. Skips spaces

monitor_readchar:
	rcall	uart_readcharwait
	mov	rPARAM1,rRETURN
	rcall	uart_writechar
	cpi	rPARAM1,' '
	breq	monitor_readchar
	ret

;-----------------------------------------------------------------
;
; monitor_write_eeprom
; Write rPARAM1 to eeprom at Z

monitor_write_eeprom:
	sbic	EECR,EEWE		;if EEWE not clear
	rjmp	monitor_write_eeprom	;    wait more

	out	EEAR,zl		; output address
	out	EEDR,rPARAM1	; output data
	
	cli			; critical section
	sbi 	EECR,EEMWE	;   set master write enable, remove if 1200 is used	
	sbi	EECR,EEWE	;   set EEPROM Write strobe
				;   This instruction takes 4 clock cycles since
				;   it halts the CPU for two clock cycles
	sei			; end critical section
	
	ret

;-----------------------------------------------------------------
;
; Constant Data

;-----------------------------------------------------------------
;
; Message

monitor_welcome:
	.db	"Monitor ready",$00
monitor_errormsg:
	.db	$0d,$0a,"Error at *",$00
monitor_exitmsg:
	.db	" OK",$00

;-----------------------------------------------------------------
