;-----------------------------------------------------------------
;
; Name:		Robot.asm
; Title:	Basic setup and services for robot.
; Version:	1.0
; Last updated:	2001.03.11
; Target:	AT90S2313
;
;-----------------------------------------------------------------
;
; DESCRIPTION
;
; Simple robot systems integration testbed.
; 
;-----------------------------------------------------------------

.nolist
.include "robotdef.inc"
.list

;-----------------------------------------------------------------
;
; Data

.dseg

robot_tickcount:
	.byte	2		; tick count

;-----------------------------------------------------------------
;
; Interrupt service vectors

.cseg
.org 0
	rjmp robot_reset	; Reset vector

	reti			; INT0 vector
	reti			; INT1 vector

	rjmp eye_capture	; T/C1 capture vector
	reti 			; T/C1 compare match vector
	rjmp eye_overflow	; T/C1 overflow vector

	rjmp robot_tick		; T/C0 overflow vector

	rjmp uart_receive	; UART receive service
	reti 			; UART data reg empty service
	reti 			; UART transmit service

	reti 			; Analog comparator service

;-----------------------------------------------------------------
;
; Start of actual code.
;
;-----------------------------------------------------------------
;
; Reset vector - generic system.
;
; Before calling reset functions for each subsystem, it sets up
; the following:
;
; 1. RAM set to all 0's
; 2. Y -> RAMSTART
; 3. Sleep mode enabled, t0, t1 on rising edge
; 4. 2kHz interrupt on timer 0
; 5. XH, YH, ZH all 0

robot_reset:
	ldi	rTEMP,RAMEND	; Set initial stack ptr location at ram end
	out	SPL,rTEMP

	clr	XH
	clr	YH
	clr	ZH

	ldi	YL,RAMEND+1	; Just past end since we are predecrementing.

robot_ramclear:
	st	-Y,YH		; clear
	cpi	YL,RAMSTART	; done?
	brne	robot_ramclear	; no.. go round again. Y-> RAMSTART at end

	; set up timer0

	ldi	rTEMP,(1<<TOIE0) ; timer interrupt mask 0 enable
	out	TIMSK,rTEMP	

	ldi	rTEMP,$02	; clock prescaler for T0 = clk/8 (02)
	out	TCCR0,rTEMP	

	; set sleep mode

	ldi	rTEMP,$2A	; enable sleep mode, interrupt t0 and t1 on falling edge only
	out	MCUCR,rTEMP

	; now call all the reset functions

	rcall	reset_uart
	rcall	reset_eyes
	rcall	reset_state
	rcall	reset_monitor

	; and, go!

	sei			; enable interrupts and off we go!
	
	ldi	rLOOPI,16
	rcall	calibrate_eyes_counted	; skip first few calibration cycles
	
	ldi	zl,low(robot_calibrate*2)
	ldi	zh,high(robot_calibrate*2)
	rcall	uart_writestringlf

	rcall	calibrate_eyes		; and then actually calibrate the eyes after we've settled

	ldi	zl,low(robot_calibrate_done*2)
	ldi	zh,high(robot_calibrate_done*2)
	rcall	uart_writestringlf

robot_mainloop:	
	sleep			; sleep till next interrupt

	rcall	update_monitor	; update monitor
	rcall	update_state	; update state machine

	rjmp	robot_mainloop	; Infinite loop - never terminates

;-----------------------------------------------------------------
;
; Timer 0 vector - free running 2kHz counter.

robot_tick:
	in	rISREGSAVE,SREG

	ldy	rISCRATCH,robot_tickcount
	inc	rISCRATCH	
	sty	robot_tickcount,rISCRATCH
	
	rcall	state_tick
	rcall	update_eyes

	out	SREG,rISREGSAVE
	reti

;-----------------------------------------------------------------
;
; Messages

robot_calibrate:
	.db	"Calibrating eyes...",$00

robot_calibrate_done:
	.db	"...done",$00

;-----------------------------------------------------------------
;
; Include other subsystems

.include "eyes.asm"
.include "uart.asm"
.include "state.asm"
.include "monitor.asm"

;-----------------------------------------------------------------

