;-----------------------------------------------------------------
;
; Name:		state.asm
; Title:	State machine for running the robot
; Version:	1.0
; Last updated:	2001.03.17
; Target:	AT90S2313
;
;-----------------------------------------------------------------
;
; DESCRIPTION
;
; The high-level behavior of the robot is determined by a state
; machine, which specifies the response to several stimuli.
;
; The states are written into the EEPROM, either by serial
; programming, or via the separate robot monitor subsystem.
;
; Each state is stored as a compressed description, as follows:
;
; Byte 0:
;	Bits 7-4, motor output. 00 = stop, 01=reverse,
;		  10=forward, 11=random state, for each motor
;	Bit  3	  reserved
;	Bits 0-2, timeout value, approximately...
;		  0000 = 1/8 second
;		  0001 = 1/4 second
;		  0010 = 1/2 second
;		  0011 =   1 second
;		  0100 =   2 seconds
;		  0101 =   4 seconds
;		  0110 =   8 seconds
;		  0111 =  16 seconds
;
; Byte 1..n:
;	Bits 7-4: Destination state for event 2n-2
;	Bits 3-0: Destination state for event 2n-1
;	Destination states are specified relative to this state:
;	0000 = +0 (self transition)
;	0001 = +1
;	...
;	0111 = +7
;	1000 = +0 (no transition - event inactive)
;	1001 = -7
;	...
;	1111 = -1
;
; Initial state is always state 0
; States are numbered 0-9, A-Z for convenience.
; Maximum number of states is set by available EEPROM space (128B)
; 
;-----------------------------------------------------------------

;;;; .include "robotdef.inc"

;-----------------------------------------------------------------
;
; Defines

.equ	STATE_NUMEVENTS	=$06	; timeout, eyes1-3, bump1-2

.equ	STATE_MOTOR	=$00
.equ	STATE_RESPONSE	=$01
.equ	STATE_SIZE	=$01+(STATE_NUMEVENTS+1)/2

.equ	STATE_COUNT	=(128/STATE_SIZE)
.equ	STATE_SPACE	=(STATE_SIZE*STATE_COUNT)

.equ	STATE_mMOTORL	=(mMOTORLA|mMOTORLB)
.equ	STATE_mMOTORR	=(mMOTORRA|mMOTORRB)
.equ	STATE_mMOTORS	=mMOTORS
.equ	STATE_mTIMEOUT	=$07

.equ	mEVENT_TIMEOUT	=$01
.equ	mEVENT_BUMPF	=$02
.equ	mEVENT_BUMPB	=$04
.equ	mEVENT_EYEL	=$08
.equ	mEVENT_EYER	=$10
.equ	mEVENT_EYEB	=$20

;-----------------------------------------------------------------
;
; Macros
;
; Multiply register by state size.
; May use rSCRATCH as temporary register
; use: mulss rN
; rN <- rN * STATE_SIZE

.macro	mulss
	lsl	@0
	lsl	@0
.endmacro

;-----------------------------------------------------------------
;
; Variables

.dseg

state_current:	.byte	1	; pointer to current state block
state_ticklo:	.byte	1	; tick count for timeout
state_tickhi:	.byte	1	; tick count for timeout
state_events:	.byte	1	; bitvector for triggered events

state_randoma:	.byte	1	; state for random number generator
state_randomb:	.byte	1
state_randomc:	.byte	1

state_lastbump:	.byte	1	; last bump switch state
state_lasteye:	.byte	1	; last eye proximity state

;-----------------------------------------------------------------
;
; Code
;
; Y is assumed to point to the RAM.
; RAM is assumed to be cleared to 0's on reset.

.cseg

;-----------------------------------------------------------------
;
; reset_state
;
; Reset state machine, update random number generator, etc.

reset_state:
	ldi	rTEMP,(mEVENT_BUMPF|mEVENT_BUMPB)
	
	sty	state_lastbump,rTEMP	; and store last bump state

	;Init the random number generator 
	
	sty	state_randoma,rTEMP	;since a 00,00,00 state will not 
	sty	state_randomb,rTEMP	;progress.
	sty	state_randomc,rTEMP	;

	; set pullups on port D for bump sensors

	sbi	PORTD,bBUMPF	
	sbi	PORTD,bBUMPB

	clr	zl
	
	rcall	state_load		; load the first state
	
	in	rTEMP2,PORTB		; but output a zero motor state to I/O port.
	cbr	rTEMP2,mMOTORS
	out	PORTB,rTEMP2
	
	ret
	
;-----------------------------------------------------------------
;
; state_load
;
; Load the current state and reinitialize state machine for
; this state. New current state pointer passed in in Z.

state_load:
	sty	state_current,zl
	
	rcall	state_readee
	mov	rTEMP,rRETURN
	mov	rTEMP2,rRETURN
	
	andi	rTEMP,STATE_mMOTORL
	cpi	rTEMP,STATE_mMOTORL
	brne	state_loadml
	
	push	rTEMP2
	rcall	state_random
	rcall	state_random
	andi	rTEMP,STATE_mMOTORL
	pop	rTEMP2
	
state_loadml:
	
	andi	rTEMP2,STATE_mMOTORR
	cpi	rTEMP2,STATE_mMOTORR
	brne	state_loadmr
	
	push	rTEMP
	rcall	state_random
	rcall	state_random
	add	rTEMP,rTEMP	; need to shift this
	add	rTEMP,rTEMP	; else we get same 2 bits as above
	andi	rTEMP,STATE_mMOTORR
	mov	rTEMP2,rTEMP
	pop	rTEMP
	
state_loadmr:
	or	rTEMP,rTEMP2
	
	; output motor state to I/O port.
	
	in	rTEMP2,PORTB
	cbr	rTEMP2,mMOTORS
	or	rTEMP2,rTEMP
	out	PORTB,rTEMP2

	; load tick count
	
	ldi	rTEMP,1
	rcall	state_readee
	mov	rTEMP2,rRETURN
	andi	rTEMP2,STATE_mTIMEOUT
	rjmp	state_loadloopstart
	
state_loadloop:
	add	rTEMP,rTEMP
	dec	rTEMP2
	
state_loadloopstart:
	brne	state_loadloop
	
	ldi	rTEMP2,$10		; leeway
	sty	state_ticklo,rTEMP2
	sty	state_tickhi,rTEMP

	ret
	
;-----------------------------------------------------------------
;
; update_state
;
; Update current state from bits in state_events. Bit 0 = event 0,
; etc. Also reads eyes.

update_state:
	rcall	read_eyes
	
	; Uncomment this block to use eye *transitions* rather than states
	; to trigger events.
	;
	; ldy	rTEMP,state_lasteye
	; sty	state_lasteye,rRETURN
	;
	; eor	rTEMP,rRETURN		; catch transitions
	
	mov	rTEMP,rRETURN
	
	lsl	rTEMP
	lsl	rTEMP
	lsl	rTEMP			; shift up to proper place
	
	ldy	rPARAM1,state_events	; get events
	or	rPARAM1,rTEMP		; or in the eyes
	sty	state_events,rPARAM1	; and store
	
	;
	; Now actually check all the bits in rPARAM1 in turn
	;
	
	ldi	rPARAM2,$01		; mask to clear bit later
	ldy	zl,state_current
	ldi	rLOOPI,STATE_NUMEVENTS-1
	
update_state_loop:
	inc	zl
	rcall	state_readee

	sbrc	rPARAM1,0
	rcall	update_state_now	; update state with HIGH nybble
	
	lsr	rPARAM1
	lsl	rPARAM2
	
	dec	rLOOPI
	brmi	update_state_exit

	swap	rRETURN

	sbrc	rPARAM1,0
	rcall	update_state_now	; update state with HIGH nybble

	lsr	rPARAM1
	lsl	rPARAM2

	dec	rLOOPI
	brpl	update_state_loop

update_state_exit:
	ret
	
update_state_now:
	; update state according to high nybble of rRETURN.	
	
	; clear event bit
	
	ldy	rTEMP,state_events
	com	rPARAM2
	and	rTEMP,rPARAM2
	com	rPARAM2
	sty	state_events,rTEMP

	; is this a null transition? if so, continue.

	mov	rTEMP,rRETURN
	andi	rTEMP,$F0
	cpi	rTEMP,$80
	breq	update_state_notrans	; null transition - skip
	
	; otherwise, we have a real event causing a transition
	
	ldy	zl,state_current	; get current state for later on, in case we do a self transition
	mov	rLOOPJ,rLOOPI		; save this for later too

	clr	rLOOPI			; no more loops
	
	cpi	rTEMP,$00
	breq	update_state_selftrans	; self transition - short cut, no uart output

	; sign extend difference
	
	asr	rTEMP
	asr	rTEMP
	asr	rTEMP
	asr	rTEMP
	
	mulss	rTEMP		; get difference
	add	zl,rTEMP	; and add
	brpl	update_state_nneg
	
	subi	zl,-STATE_SPACE	; adjust if -ve.
	rjmp	state_load
	
update_state_nneg:
	cpi	zl,STATE_SPACE
	brcs	update_state_novr
	
	subi	zl,STATE_SPACE
	
update_state_novr:

	;
	; Write triggered event to UART, and continue
	;
	
	push	zl
	ldy	rPARAM1,state_current
	rcall	monitor_statetochar
	rcall	uart_writechar

	ldi	rPARAM1,':'
	rcall	uart_writechar

	ldi	rPARAM1,STATE_NUMEVENTS-1+'0'
	sub	rPARAM1,rLOOPJ
	rcall	uart_writechar

	ldi	rPARAM1,'-'
	rcall	uart_writechar
	ldi	rPARAM1,'>'
	rcall	uart_writechar
	mov	rPARAM1,zl
	rcall	monitor_statetochar
	rcall	uart_writechar

	rcall	uart_writecrlf
	pop	zl

update_state_selftrans:
	rjmp	state_load

update_state_notrans:
	; push	rPARAM1
	; rcall	uart_writecrlf
	; pop	rPARAM1
	ret
		
;-----------------------------------------------------------------
;
; state_tick
;
; Tick counter for state machine timeout values, etc.

state_tick:
	ldy	rISCRATCH,state_ticklo
	dec	rISCRATCH
	sty	state_ticklo,rISCRATCH

	brne	state_tickend
	
	; 1/8 s basic tick. Sample the bump switches.
	
	in	rITEMP2,GIFR		; read IFR for bumps
	out	GIFR,rITEMP2		; and reset set bits
	
	; uncomment next block to allow events triggered on
	; bump switch TRANSITIONS, not levels.

	; ldy	rITEMP,state_lastbump
	; sty	state_lastbump,rITEMP2
	
	; eor	rITEMP2,rITEMP		; catch TRANSITIONS not LEVELS
	
	swap	rITEMP2
	lsr	rITEMP2
	andi	rITEMP2,(mEVENT_BUMPF|mEVENT_BUMPB)

	ldy	rITEMP,state_events
	or	rITEMP,rITEMP2		; indicate events
	
	; now check for timeout
	
	ldy	rISCRATCH,state_tickhi
	dec	rISCRATCH

	sty	state_tickhi,rISCRATCH

	brne	state_notimeout
	
	ori	rITEMP,mEVENT_TIMEOUT	; indicate events
	
state_notimeout:
	sty	state_events,rITEMP	; and put event status back
	
state_tickend:
	ret

;-----------------------------------------------------------------
;
; state_readee
;
; Read from eeprom, address Z, return in rRETURN

state_readee:
	sbic	EECR,EEWE	; check not writing
	rjmp	state_readee
	
	out	EEAR,zl
	sbi	EECR,EERE
	in	rRETURN,EEDR
	ret

;-----------------------------------------------------------------
;
; state_random
;
; Random number generator
; Maximal length 19 bit shift register sequence, adapted from the
; "getting started" notes of Dave van Horn. Thanks, Dave!

state_random:
	;  rSCRATCH3    rSCRATCH2    rSCRATCH
	;22222111 11111110 00000000
	;43210987 65432109 87654321
	
	ldy	rSCRATCH,state_randoma
	ldy	rSCRATCH2,state_randomb
	ldy	rSCRATCH3,state_randomc

	mov	rTEMP,rSCRATCH		;Make copy
	mov	rTEMP2,rSCRATCH3	;Make copy

	rol	rSCRATCH		;Shift the bits D7->Carry
	rol	rSCRATCH2		;Carry->D0 D7->Carry
	rol	rSCRATCH3		;Carry->D0 D7->Carry, but that will be fixed.

	;Now we have to Xor the bits to see what
	;goes in to bit 1

	rol	rTEMP2			;Move bit 19->20
	andi	rTEMP2,$08		;Important that D1,0 be zero

	andi	rTEMP,$13		;Mask out irrelevants, protecting bit 19
	or	rTEMP2,rTEMP		;Get bits 5->21 2->18 1->17

	andi	rTEMP2,$18		;Nuke bits 18,17
	lsr	rTEMP2			;19->19 5->20
	lsr	rTEMP2			;19->18 5->19
	lsr	rTEMP2			;19->17 5->18

	eor	rTEMP2,rTEMP		;First xor, result in rTEMP2 (5x2 in 02 and 19x1 in 01)
	mov	rTEMP,rTEMP2		;Make a copy
	ror	rTEMP			;Move the 5x2 result to 01
	andi	rTEMP2,$01		;Mask off everything else
	andi	rTEMP,$01		;in both
	eor	rTEMP2,rTEMP		;
	brne	state_random_1		;If one, do that, else 

state_random_0:	
	;Set a zero in the lsb of the low byte
	mov	rTEMP,rSCRATCH		;Get the low byte
	andi	rTEMP,$FE		;Make the LSB zero
	mov	rSCRATCH,rTEMP		;Put it back
	rjmp	state_random_exit	;Bye bye

state_random_1:	
	;Set a one in the lsb of the low byte
	mov	rTEMP,rSCRATCH		;Get the low byte
	ori	rTEMP,$01		;Make the LSB one
	mov	rSCRATCH,rTEMP		;Put it back

state_random_exit:
	sty	state_randoma,rSCRATCH
	sty	state_randomb,rSCRATCH2
	sty	state_randomc,rSCRATCH3

	mov	rRETURN,rSCRATCH	; and return it

	ret

;-----------------------------------------------------------------

;-----------------------------------------------------------------
;
; Default state machine
; Derived from program loaded into monitor as in the comments

.eseg
.org	0
				;  +------------------------ state number
				;  |  ++-------------------- motor state
				;  |  ||  +----------------- timeout value
				;  |  ||  |                  Next state...
				;  |  ||  |   +------------- ...on timeout
				;  |  ||  |   | ++---------- ...on bump switches, fore or back
				;  |  ||  |   | || +++------ ...on eye activation, left, right or back
				;  |  ||  |   | || |||
	.db 	$F5,$13,$25,$42	; s0  ??  5   1 32 542     ; exploring  4s, run away
	.db	$F5,$62,$14,$31	; s1  ??  5   7 32 542     ; exploring  4s, run away
	.db	$93,$E1,$83,$28	; s2  FF  3   0 3- 54-     ; shy forward
	.db	$63,$D8,$F8,$8F	; s3  RR  3   0 -2 --2     ; shy backward
	.db 	$23,$FF,$EF,$8E	; s4  R0  3   3 32 3-2     ; reverse left
	.db	$43,$EE,$D8,$ED	; s5  0R  3   3 32 -32     ; reverse right
	.db	$F5,$AD,$CF,$EC	; s6  ??  5   0 32 542     ; exploring  4s, run away
	.db	$F5,$92,$14,$31	; s7  ??  5   0 98 BA8     ; exploring  4s; aggressive!
	.db	$63,$E1,$83,$28	; s8  RR  3   6 9- BA-     ; attack backward, then run away
	.db	$93,$D8,$F8,$8F	; s9  FF  3   6 -8 --8     ; attack forward, then run away
	.db	$13,$FF,$EF,$8E	; sA  F0  3   9 98 9-8     ; attack right, then forward, then run away
	.db	$83,$EE,$D8,$ED	; sB  0F  3   9 98 -98     ; attack left, then forward, then run away
	
;-----------------------------------------------------------------
	