;-----------------------------------------------------------------
;
; Name:		uart.asm
; Title:	AVR Uart services
; Version:	1.0
; Last updated:	2001.02.03
; Target:	AT90Sxxxx (All devices)
;
;-----------------------------------------------------------------
;
; DESCRIPTION
;
; UART service which uses interrupts to read characters
; from the serial port and buffer them, but which handles trans-
; mission synchronously, without interrupts.
; 
;-----------------------------------------------------------------

.dseg

uart_rhead:	.byte	1
uart_rtail:	.byte	1
uart_rsize:	.byte	1
uart_rerror:	.byte	1

.equ	UART_BUFSIZE	=$20	; must be power of 2
.equ	UART_BUFFER	=$A0	; must be on a UART_BUFSIZE boundary

.equ	UART_BAUDRATE	=9600
.equ	UART_BAUDK	=(ROBOT_CLOCKRATE/(16*UART_BAUDRATE))-1

;-----------------------------------------------------------------

.cseg

;-----------------------------------------------------------------
;
; Reset vector - setup

reset_uart:
	ldi	rTEMP,UART_BUFFER
	sty	uart_rhead,rTEMP
	sty	uart_rtail,rTEMP

	ldi	rTEMP,UART_BAUDK
	out	UBRR,rTEMP	; load baudrate
	
	ldi	rTEMP,(1<<TXEN)|(1<<RXEN)|(1<<RXCIE)	; enable transmit/receive, enable rx int.
	out	UCR,rTEMP

	ldi	ZH,high(Message*2)	; haha! Word address
	ldi	ZL,low(Message*2)
	rcall	uart_writestringlf

	clr	ZH
	ret

;-----------------------------------------------------------------
;
; Receive vector - buffer input

uart_receive:
	in	rISREGSAVE,SREG
	mov	rISCRATCH4,XL

	sbic	USR,FE		; check for
	rjmp	uart_rxerr_fe	; framing error

	ldy	rITEMP,uart_rsize
	cpi	rITEMP,UART_BUFSIZE
	breq	uart_rxerr_bf	; buffer full

	inc	rITEMP
	sty	uart_rsize,rITEMP

	ldy	XL,uart_rhead
	in	rITEMP,UDR
	st	X+,rITEMP

	andi	XL,(UART_BUFSIZE-1)
	ori	XL,UART_BUFFER
	sty	uart_rhead,XL

	sbic	USR,OR		; check for
	rjmp	uart_rxerr_or	; overrun

uart_rxend:
	mov	XL,rISCRATCH4
	out	SREG,rISREGSAVE
	reti

uart_rxerr_fe:
uart_rxerr_bf:
	in	rITEMP,UDR		; dummy read to clear RxC
uart_rxerr_or:
	ldi	rITEMP,1		; overflow, framing error
	sty	uart_rerror,rITEMP

	rjmp	uart_rxend

;-----------------------------------------------------------------
;
; Read from receiver buffer char=rRETURN, wait until character ready

uart_readcharwait:
	rcall	uart_charready
	breq	uart_readcharwait

uart_readchar:
	push	XL
	
	ldy	XL,uart_rtail
	ld	rRETURN,x+
	andi	XL,(UART_BUFSIZE-1)
	ori	XL,UART_BUFFER
	sty	uart_rtail,XL

	cli
	ldy	rSCRATCH,uart_rsize
	dec	rSCRATCH
	sty	uart_rsize,rSCRATCH
	sei

	pop	XL

	ret
	
;-----------------------------------------------------------------
;
; Ask if a character is ready from the UART (eq=no, ne=yes)

uart_charready:
	ldy	rSCRATCH,uart_rsize
	and	rSCRATCH,rSCRATCH
	ret	

;-----------------------------------------------------------------
;
; Write to transmitter a string followed by a cr/lf

uart_writestringlf:
	rcall	uart_writestring

	; falls through to uart_writecrlf...

;-----------------------------------------------------------------
;
; Write to transmitter a cr/lf pair

uart_writecrlf:
	ldi	rPARAM1,$0D
	rcall	uart_writechar
	ldi	rPARAM1,$0A

	; falls through to uart_writechar again...

;-----------------------------------------------------------------
;
; Write to transmitter char=rPARAM1

uart_writechar:
	sbis	USR,UDRE
	rjmp	uart_writechar
	out	UDR,rPARAM1
	ret

;-----------------------------------------------------------------
;
; Write to transmitter hex value=rPARAM1

uart_writehex:
	push	rPARAM1
	swap	rPARAM1
	rcall	uart_writehexit
	pop	rPARAM1
	
uart_writehexit:
	andi	rPARAM1,$0F
	subi	rPARAM1,-48
	cpi	rPARAM1,$3A
	brcs	uart_writechar
	subi	rPARAM1,-7
	rjmp	uart_writechar

;-----------------------------------------------------------------
;
; Write 0-terminated string in Flash at Z to transmitter

uart_writestring:
	lpm				; r0,Z
	adiw	ZL,1
	mov	rPARAM1,r0
	and	rPARAM1,rPARAM1
	breq	uart_writestringend

	rcall	uart_writechar

	rjmp	uart_writestring

uart_writestringend:
	ret				; NB: Shared with uart_writestring
	
;-----------------------------------------------------------------
;
; Write 0-terminated string in RAM at Z to transmitter, with crlf
; if reqd.
;
; - - - UNUSED - - - - - - - - - - - - - - - - - - - - - - - - - -
;
;uart_writestringram:
;	clr	zh
;uart_writestringramloop:
;	ld	rPARAM1,z+
;	and	rPARAM1,rPARAM1
;	breq	uart_writestringend
;
;	rcall	uart_writechar
;
;	rjmp	uart_writestringramloop
;
;uart_writestringramlf:
;	rcall	uart_writestringram
;	rjmp	uart_writecrlf
;
; - - - UNUSED - - - - - - - - - - - - - - - - - - - - - - - - - -

;-----------------------------------------------------------------
;
; Constant Data

;-----------------------------------------------------------------
;
; Message

Message:
	.db	$0d,$0a,"UART Ready",$00

;-----------------------------------------------------------------
